// written by Zheng Gengbin
// pcrcFortran functions call maker
// and some facilities.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>

#include "compilerApi.h"
#include "strategy.h"

#include "anls_att.h"

//extern "C" int strcasecmp(...);

extern char *func_name_search(SgFunctionCallExp *);


// flag 0: FUNCTION_NAME 1:PROCEDURE_NAME
void replace_intrinsic_function_name(SgExpression *proc_call,int flag)
{
	SgFunctionCallExp *call = (SgFunctionCallExp *)proc_call;
	char *name=func_name_search((SgFunctionCallExp *)call);
if (!silentFlag)
printf("Intrinsic name:%s\n",name);
	if (strlen(name) != 0 )
        {
                SgSymbol *newsym = new SgSymbol(flag==0?FUNCTION_NAME:PROCEDURE_NAME,name);
                newsym->thesymb->attr = call->funName()->attributes();
                call->setSymbol(*newsym);
		delete newsym;
        }
}

// change the function call to procedure call 
// the function call is now unified to : A = func (X)
void intrinsic2sub(SgStatement *s)
{
	  // check if it can be converted
	SgExpression *lhs = s->expr(0);
	SgExpression *rhs = s->expr(1);
	SgFunctionCallExp *funccall = isSgFunctionCallExp(rhs);
	if ( funccall == NULL ) return;
	if (! isSgIntrinsicFunctionCallExp(rhs) && ! isSgArrayType(lhs->type())) return;

	  // if is scalar type , and the function is fortran77 compatible ,return
	SgArrayType *returnedType = isSgArrayType(funccall->type());
	if (returnedType==NULL && ((SgFunctionSymb*)funccall->symbol())->is_elementary_function() ) return ;
	if (funccall->variant() != INTRIN_FUNC_CALL) return;
	  // if this intrinsic function is deal like functions.
	if (function_intrinsic(funccall->symbol())) return;

if (!silentFlag)
printf("Intrinsic function:%s\n",funccall->symbol()->identifier());

	SgExpression *lexprlist = new SgExpression(EXPR_LIST,lhs, rhs , NULL);
	lexprlist->setRhs(funccall->lhs());
	funccall->setLhs(lexprlist);
	funccall->thellnd->variant = PROC_CALL;
	// set the BIF_LL1 & BIF_LL2 according to PROC_STAT
	s->setExpression(0,funccall);
	BIF_LL2(s->thebif) = LLNULL;
        // change the statement variant
	s->thebif->variant = PROC_STAT;
	delete lexprlist;
	s->addAttribute(INTRIN_SUB, NULL, 0);
}

/***************************************************************************
		FUNCTION CALL MAKER
***************************************************************************/

// get kind_val from T_KIND
SgExpression *get_kind_val(SgStatement *s, SgType *type)
{
    SgType *kindType = type->baseType();
    if (kindType && isSgKindType(kindType))
    {
	SgExpression *kindexp = isSgKindType(kindType)->kindVal();
	ASSERT(kindexp);
	kindexp = kindexp->lhs();
	return kindexp;
    }
    return NULL;
}

SgExpression *get_type_size(SgStatement *s, SgType *type)
{
    SgSymbol * size = s->getSymbolInCurrentScope(mixstring, sizeNameOfType(type), NULL); 
    SgExpression * sizeExp = new SgVarRefExp(size);

    SgType *baseType = type;
    if (!isSgKindType(type) && type->hasBaseType())
    	baseType=type->baseType();
    if (isSgKindType(baseType))
    {
	SgKindType *kind = isSgKindType(baseType);
	SgExpression *kindexp = kind->kindVal();
	ASSERT(kindexp);
	sizeExp = &( *sizeExp * *kindexp );
    }
    return sizeExp;
}

SgExpression *get_in_core_size(SgStatement *s)
{
  if ( have_main_program )
	return new SgValueExp(IN_CORE_SIZE);
  else {
  	SgSymbol *incoresize = s->getSymbolInCurrentScope("IN_CORE_SIZE");
	return new SgVarRefExp(incoresize);
  }
}

SgExpression *get_record_size(SgStatement *s)
{
  if ( have_main_program )
	return new SgValueExp(RECORD_SIZE);
  else {
  	SgSymbol *recordsize = s->getSymbolInCurrentScope("RECORD_SIZE");
	return new SgVarRefExp(recordsize);
  }
}

SgStatement *make_init(SgStatement *s, int store, int debugFlag)
{
    SgSymbol * fname = s->GetSymbolInScope(s, mixstring, "init", NULL);
    SgCallStmt * funcall = new SgCallStmt(*fname);

    {
      SgSymbol *storageSym;
      switch (store) {
      case COLUMN_MAJOR:
    	storageSym = s->getSymbolInCurrentScope("COLUMN_MAJOR"); break;
      case ROW_MAJOR:
    	storageSym = s->getSymbolInCurrentScope("ROW_MAJOR"); break;
      }
      SgExpression *storageExp = new SgVarRefExp(storageSym);
      funcall->addArg(*storageExp);
      delete storageExp;
    }

    SgExpression *deb = new SgValueExp(debugFlag);
    funcall->addArg(*deb);    

    return funcall;
}

//
// rank: default:-1
// grp:  default:NULL
SgStatement *make_new_array_data(SgStatement *s,SgSymbol *dad,SgType *arraysymbtype,int store,int rank,SgExpression *grpExp)
{
    SgSymbol * symb = s->getSymbolInCurrentScope(mixstring, "new_array_data", NULL);
    SgFunctionRefExp * funcall = new SgFunctionRefExp (*symb);

#ifdef MEM_ALLOC
    SgSymbol * stack = s->getSymbolInCurrentScope(stackNameOfType(arraysymbtype),"mb", NULL); 
#else
    SgSymbol * stack = s->getSymbolInCurrentScope(mixstring, stackNameOfType(arraysymbtype), NULL); 
#endif
    SgExpression * stackExp = new SgVarRefExp(stack);
    funcall->addArg(*stackExp);    
						       
    SgSymbol * nametype = s->getSymbolInCurrentScope(mixstring, nameOfType(arraysymbtype), NULL); 
    SgExpression * nametypeExp = new SgVarRefExp(nametype);
    funcall->addArg(*nametypeExp);    
						       
    SgExpression * sizeExp = get_type_size(s, arraysymbtype);
    funcall->addArg(*sizeExp);    
						       
    SgValueExp *v0;
    if (rank == -1) 
    {
      if (isSgArrayType(arraysymbtype))
    	v0 = new SgValueExp(((SgArrayType *)arraysymbtype)->dimension());
      else
    	v0 = new SgValueExp(0);
    }
    else
    	v0 = new SgValueExp(rank);
    funcall->addArg(*v0);    
   
    if (grpExp==NULL) 
	grpExp = new SgVarRefExp(s->getSymbolInCurrentScope(mixstring, "grp_glob", NULL)); 
    funcall->addArg(*grpExp);    

    {
      SgSymbol *storageSym;
      if (store == -1) store = storageMode;
      switch (store) {
      case COLUMN_MAJOR:
    	storageSym = s->getSymbolInCurrentScope("COLUMN_MAJOR"); break;
      case ROW_MAJOR:
    	storageSym = s->getSymbolInCurrentScope("ROW_MAJOR"); break;
      }
      SgExpression *storageExp = new SgVarRefExp(storageSym);
      funcall->addArg(*storageExp);
    }

    SgExpression * dadBaseExp=new SgVarRefExp(dad);

    SgAssignStmt * assignStmt = new SgAssignStmt(*dadBaseExp, *funcall);

    delete dadBaseExp, grpExp, v0, sizeExp, nametypeExp, stackExp, funcall;
  
    return assignStmt;
}

// rank: default:-1
// grpExp:  default:NULL
SgStatement *make_new_array_data1(SgStatement *s,SgSymbol *dad,SgExpression *address,SgType *arraysymbtype,int store,int rank,SgExpression *grpExp)
{
    SgSymbol * symb = s->getSymbolInCurrentScope(mixstring, "new_array_data", NULL);
    SgFunctionRefExp * funcall = new SgFunctionRefExp (*symb);

    funcall->addArg(*address);    
						       
    SgSymbol * nametype = s->getSymbolInCurrentScope(mixstring, nameOfType(arraysymbtype), NULL); 
    SgExpression * nametypeExp = new SgExpression(VAR_REF, NULL, NULL, nametype);
    funcall->addArg(*nametypeExp);    
						       
    SgExpression * sizeExp = get_type_size(s, arraysymbtype);
    funcall->addArg(*sizeExp);    
						       
    SgValueExp *v0;
    if (rank == -1) 
    {
      if (isSgArrayType(arraysymbtype))
    	v0 = new SgValueExp(((SgArrayType *)arraysymbtype)->dimension());
      else
    	v0 = new SgValueExp(0);
    }
    else
    	v0 = new SgValueExp(rank);
    funcall->addArg(*v0);    
   
    if (grpExp==NULL) 
	grpExp = new SgVarRefExp(s->getSymbolInCurrentScope(mixstring, "grp_glob", NULL)); 
    funcall->addArg(*grpExp);    

    {
      SgSymbol *storageSym;
      if (store == -1) store = storageMode;
      switch (store) {
      case COLUMN_MAJOR:
    	storageSym = s->getSymbolInCurrentScope("COLUMN_MAJOR"); break;
      case ROW_MAJOR:
    	storageSym = s->getSymbolInCurrentScope("ROW_MAJOR"); break;
      }
      SgExpression *storageExp = new SgVarRefExp(storageSym);
      funcall->addArg(*storageExp);
    }

    SgExpression * dadBaseExp=new SgVarRefExp(dad);

    SgAssignStmt * assignStmt = new SgAssignStmt(*dadBaseExp, *funcall);
  
    delete funcall, nametypeExp, sizeExp, v0, grpExp, dadBaseExp;

    return assignStmt;
}

SgStatement *make_set_type(SgStatement *s,SgSymbol *dad, SgType *arraysymbtype)
{
    SgSymbol * symb = s->getSymbolInCurrentScope(mixstring, "set_type", NULL);
    SgCallStmt * funcall = new SgCallStmt (*symb);

    SgExpression * dadBaseExp=new SgVarRefExp(dad);
    funcall->addArg(*dadBaseExp);    
    delete dadBaseExp;

    SgSymbol * nametype = s->getSymbolInCurrentScope(mixstring, nameOfType(arraysymbtype), NULL); 
    SgExpression * nametypeExp = new SgVarRefExp(nametype);
    funcall->addArg(*nametypeExp);    
    delete nametypeExp;

    SgExpression * sizeExp = get_type_size(s, arraysymbtype);
    funcall->addArg(*sizeExp);    

    return funcall ;
}

SgStatement * make_new_array_copy(SgStatement *s,SgSymbol *dest,SgSymbol *src)
{
    SgSymbol * symb = s->getSymbolInCurrentScope(mixstring, "new_array_copy", NULL);
    SgFunctionRefExp * funcall = new SgFunctionRefExp (*symb);

    SgType *arraysymbtype = src->type();

#ifdef MEM_ALLOC
    SgSymbol * stack = s->getSymbolInCurrentScope(stackNameOfType(arraysymbtype),"mb", NULL); 
#else
    SgSymbol * stack = s->getSymbolInCurrentScope(mixstring, stackNameOfType(arraysymbtype), NULL); 
#endif
    SgExpression * stackExp = new SgVarRefExp(stack);
    funcall->addArg(*stackExp);    
						       
    SgExpression * srcdadExp=new SgVarRefExp(src);
    funcall->addArg(*srcdadExp);    

    SgExpression * destdadExp=new SgVarRefExp(dest);

    SgAssignStmt * assignStmt = new SgAssignStmt(*destdadExp, *funcall);
    return assignStmt;
}

SgStatement * make_set_array_copy(SgStatement *s, SgSymbol *dad, int dim, SgSymbol *dad_parent, int dim_parent)
{
	SgSymbol * symb=s->getSymbolInCurrentScope(mixstring, "set_array_copy", NULL);
	SgCallStmt * funcall= new SgCallStmt(*symb);

	SgExpression *dadexp = new SgVarRefExp(dad);
	funcall->addArg(*dadexp);
	delete dadexp;

	SgExpression *dimexp = new SgValueExp(dim);
	funcall->addArg(*dimexp);
	delete dimexp;

	dimexp = new SgValueExp(dim_parent);
	SgExpression *range_exp = make_range(s,dad_parent,dimexp);
	funcall->addArg(*range_exp);
	delete dimexp;

	return funcall;
}

SgStatement * make_set_array_collapse(SgStatement *s, SgSymbol *dad, int dim, SgExpression *lb, SgExpression *ub)
{
	SgSymbol * symb=s->getSymbolInCurrentScope(mixstring, "set_array_collapse", NULL);
	SgCallStmt * funcall= new SgCallStmt(*symb);

	SgExpression *dadexp = new SgVarRefExp(dad);
	funcall->addArg(*dadexp);
	delete dadexp;

	SgExpression *dimexp = new SgValueExp(dim);
	funcall->addArg(*dimexp);
	delete dimexp;

	funcall->addArg(*lb);
	funcall->addArg(*ub);
	return funcall;
}

SgStatement * make_set_array_distribute(SgStatement *s, SgSymbol *dad, int dim, SgExpression *start, SgExpression *end, int distcode, SgSymbol *procgrd, int processorDimCounter)
{
	SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "set_array_distribute",NULL);
      	SgCallStmt * afuncall = new SgCallStmt(*fname);

        SgExpression * expr = new SgVarRefExp(dad);
     	afuncall->addArg(*expr);

      	SgValueExp * dimexp = new SgValueExp(dim);
      	afuncall->addArg(*dimexp);

      	afuncall->addArg(*start);
      	afuncall->addArg(*end);

      	SgExpression * distExp = new SgValueExp(distcode);
      	afuncall->addArg(*distExp);

	SgExpression * procgrdexpr = new SgVarRefExp(procgrd);
	afuncall->addArg(*procgrdexpr);
	    
	SgValueExp * pdimexp = new SgValueExp(processorDimCounter);	      
	afuncall->addArg(*pdimexp);

	return afuncall;
}

SgStatement * make_set_array_align(SgStatement *s, SgSymbol *dad, int dim, SgExpression *start, SgExpression *end, SgExpression *a ,SgExpression *b, SgExpression *gl, SgExpression *gu, SgExpression  *rng)
{
      SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "set_array_align", NULL);
      SgCallStmt * afuncall = new SgCallStmt(*fname);

      SgExpression * expr = new SgVarRefExp(dad);
      afuncall->addArg(*expr);
      delete expr;

      SgValueExp * dimexp = new SgValueExp(dim);
      afuncall->addArg(*dimexp);
      delete dimexp;

      afuncall->addArg(*start);
      afuncall->addArg(*end);

      afuncall->addArg(*a);
      afuncall->addArg(*b);

      afuncall->addArg(*gl);
      afuncall->addArg(*gu);

      afuncall->addArg(*rng);

      return afuncall;
}
     
SgStatement * make_remap(SgStatement *s, SgSymbol *dadsec, SgSymbol *dadsrc)
{
	SgSymbol * symb=s->getSymbolInCurrentScope(mixstring, "remap", NULL);
	SgCallStmt * funcall= new SgCallStmt(*symb);
	SgExpression * dadexp = new SgVarRefExp(dadsec);
	funcall->addArg(*dadexp);
	delete dadexp;
	dadexp = new SgVarRefExp(dadsrc);
	funcall->addArg(*dadexp);
	delete dadexp;
	return  funcall;
}

SgStatement * make_write_halo(SgStatement * s, SgSymbol *dad,SgSymbol *gtl, SgSymbol *gtu)
{
  SgExpression * dadExp = new SgVarRefExp(dad);
  
  SgExpression * gtlarray = new SgArrayRefExp(*gtl);

  SgExpression * gtuarray = new SgArrayRefExp(*gtu);
  
  SgSymbol * writeHalo=s->getSymbolInCurrentScope(mixstring, "write","halo");
  SgCallStmt * func = new SgCallStmt(*writeHalo);

  func->addArg(*dadExp);
  func->addArg(*gtlarray);
  func->addArg(*gtuarray);
  
  return func;
}

SgStatement * make_write_back_halo(SgStatement * s, SgSymbol *dad,SgSymbol *gtl, SgSymbol *gtu)
{
  SgExpression * dadExp = new SgVarRefExp(dad);
  
  SgExpression * gtlarray = new SgArrayRefExp(*gtl);

  SgExpression * gtuarray = new SgArrayRefExp(*gtu);
  
  SgSymbol * writeHalo=s->getSymbolInCurrentScope(mixstring, "write_back","halo");
  SgCallStmt * func = new SgCallStmt(*writeHalo);

  func->addArg(*dadExp);
  func->addArg(*gtlarray);
  func->addArg(*gtuarray);
  
  return func;
}

SgStatement * make_delete_array(SgStatement *s , SgSymbol *dad)
{
    SgSymbol * symb = s->getSymbolInCurrentScope(mixstring, "delete_array", NULL);
    SgCallStmt * funcall = new SgCallStmt (*symb);

    SgExpression * baseexpr=new SgVarRefExp(dad);
    funcall->addArg(*baseexpr);
    delete baseexpr;
    return funcall;
}

SgStatement * make_array_free(SgStatement *s , SgSymbol *dad)
{
    SgSymbol * symb = s->getSymbolInCurrentScope(mixstring, "array_free", NULL);
    SgCallStmt * funcall = new SgCallStmt (*symb);

    SgExpression * baseexpr=new SgVarRefExp(dad);
    funcall->addArg(*baseexpr);
    delete baseexpr;
    return funcall;
}

SgExpression *make_new_group_grid(SgStatement *s,int dim, SgSymbol *shp)
{
      SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "new_group_grid", NULL);
      SgFunctionRefExp * funcexpr = new SgFunctionRefExp (*fname);

      SgValueExp * dimExp = new SgValueExp(dim);
      funcexpr->addArg(*dimExp);
      delete dimExp;

      SgExpression * procshp= new SgVarRefExp(shp);
      funcexpr->addArg(*procshp);
      delete procshp;
      return funcexpr;
}

SgStatement * make_delete_group(SgStatement *s , SgSymbol *grp)
{
      SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "delete_group", NULL);

      SgCallStmt * afuncall = new SgCallStmt(*fname);

      SgExpression *grpexp = new SgVarRefExp(grp);
      afuncall->addArg(*grpexp);
      return afuncall;
}

SgExpression * make_new_range_distribute(SgStatement *s,SgExpression *start,SgExpression *end,int distcode,SgSymbol *grp,int dim)
{
	SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "new_range_distribute", NULL);
	SgFunctionRefExp * funcexpr = new SgFunctionRefExp (*fname);
	    
	funcexpr->addArg(*start);
	    
	funcexpr->addArg(*end);

	// distributing code;
	//SgValueExp * dist = new SgValueExp(distcode);
	SgSymbol *distSymb;
	switch (distcode)
	{
	case 1:
	     distSymb = s->getSymbolInCurrentScope(mixstring, "block", NULL);
	     break;
	case 2:
	     distSymb = s->getSymbolInCurrentScope(mixstring, "cyclic", NULL);
	     break;
	default:
	     ASSERT(0);
	}
	SgExpression *dist = new SgVarRefExp(distSymb);
	funcexpr->addArg(*dist);

	// in the following we want the grp name of the prcossor;

	SgExpression * procgrdexpr = new SgVarRefExp(grp);
	funcexpr->addArg(*procgrdexpr);
	    
	SgValueExp * dimExp = new SgValueExp(dim);	      
	funcexpr->addArg(*dimExp);

	return funcexpr;
}

SgExpression *make_new_range_collapse(SgStatement *s,SgExpression *start,SgExpression *end)
{
	SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "new_range_collapse", NULL);
	SgFunctionRefExp * funcexpr = new SgFunctionRefExp (*fname);
	    
	funcexpr->addArg(*start);
	funcexpr->addArg(*end);

	return funcexpr;
}

SgStatement * make_delete_range(SgStatement *s , SgExpression *rngExp)
{
      SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "delete_range", NULL);
      SgCallStmt * afuncall = new SgCallStmt(*fname);
      afuncall->addArg(*rngExp);
      return afuncall;
}

// base = pcrc_integer_alloc (pcrc_size (pcrc_tdd0))
SgStatement * make_alloc(SgStatement *s,SgType *type,SgSymbol *base,SgSymbol *dad)
{
  // allocate memory;
    SgExpression * baseexp=new SgVarRefExp(base);

    SgSymbol * franame = s->getSymbolInCurrentScope(mixstring, allocNameOfType(type), NULL);
    SgFunctionRefExp * rafuncexpr = new SgFunctionRefExp (*franame);

    SgExpression *sizefuncexpr = make_in_core_size(s, dad);

    SgExpression *kindexp = get_kind_val(s, type);
    if (kindexp)
	sizefuncexpr = & (*sizefuncexpr * *kindexp);

    rafuncexpr->addArg(*sizefuncexpr);

    SgStatement * alloc = new SgAssignStmt(*baseexp, *rafuncexpr);

    delete baseexp, rafuncexpr, sizefuncexpr;

    return alloc;
}

// base = pcrc_integer_stack_alloc (pcrc_in_core_size (pcrc_tdd0),pcrc_mh?)
SgStatement * make_stack_alloc(SgStatement *s,SgType *type,SgSymbol *base,SgSymbol *dad,SgSymbol * mHandle)
{
  // allocate memory;
    SgExpression * baseexp=new SgVarRefExp(base);

    SgSymbol * franame = s->getSymbolInCurrentScope(mixstring, allocNameOfType(type), NULL);
    SgFunctionRefExp * rafuncexpr = new SgFunctionRefExp (*franame);

    SgExpression *sizefuncexpr = make_in_core_size(s, dad);

    SgExpression *kindexp = get_kind_val(s, type);
    if (kindexp)
	sizefuncexpr = & (*sizefuncexpr * *kindexp);

    rafuncexpr->addArg(*sizefuncexpr);

    SgExpression * mhexp=new SgVarRefExp(mHandle);
    rafuncexpr->addArg(*mhexp);

    SgStatement * alloc = new SgAssignStmt(*baseexp, *rafuncexpr);

    delete baseexp, rafuncexpr, sizefuncexpr, mhexp;

    return alloc;
}

// base = pcrc_integer_stack_alloc (size,pcrc_mh?)
SgStatement * make_stack_alloc_size(SgStatement *s,SgType *type,SgSymbol *base,SgExpression *sizeexp,SgSymbol * mHandle)
{
  // allocate memory;
    SgExpression * baseexp=new SgVarRefExp(base);

    SgSymbol * franame = s->getSymbolInCurrentScope(mixstring, allocNameOfType(type), NULL);
    SgFunctionRefExp * rafuncexpr = new SgFunctionRefExp (*franame);

    rafuncexpr->addArg(*sizeexp);

    SgExpression * mhexp=new SgVarRefExp(mHandle);
    rafuncexpr->addArg(*mhexp);

    SgStatement *alloc = new SgAssignStmt(*baseexp, *rafuncexpr);

    delete baseexp, rafuncexpr, mhexp;

    return alloc;
}

// call pcrc_mem_free
SgStatement * make_free(SgStatement *s,SgType *type,SgSymbol *base)
{
    // call real free;
    SgSymbol * symb = s->getSymbolInCurrentScope(mixstring, freeNameOfType(type), NULL);
    SgCallStmt * funcall = new SgCallStmt (*symb);
    
    SgExpression * baseexpr=new SgVarRefExp(base);
    
    funcall->addArg(*baseexpr);

    delete baseexpr;

    return funcall;
}

// base = pcrc_integer_heap_alloc (pcrc_in_core_size (pcrc_tdd0),pcrc_mh?)
SgStatement * make_heap_alloc(SgStatement *s,SgType *type,SgSymbol *base,SgSymbol *dad,SgSymbol * mHandle)
{
  // allocate memory;
    SgExpression * baseexp=new SgVarRefExp(base);

    SgSymbol * franame = s->getSymbolInCurrentScope(mixstring, allocHeapNameOfType(type), NULL);
    SgFunctionRefExp * rafuncexpr = new SgFunctionRefExp (*franame);

    SgExpression *sizefuncexpr = make_in_core_size(s, dad);

    SgExpression *kindexp = get_kind_val(s, type);
    if (kindexp)
	sizefuncexpr = & (*sizefuncexpr * *kindexp);

    rafuncexpr->addArg(*sizefuncexpr);

    SgExpression * mhexp=new SgVarRefExp(mHandle);
    rafuncexpr->addArg(*mhexp);

    SgStatement * alloc = new SgAssignStmt(*baseexp, *rafuncexpr);

    delete baseexp, rafuncexpr, sizefuncexpr, mhexp;

    return alloc;
}

// base = pcrc_integer_heap_alloc (size,pcrc_mh?)
SgStatement * make_heap_alloc_size(SgStatement *s,SgType *type,SgSymbol *base,SgExpression *sizeexp,SgSymbol * mHandle)
{
  // allocate memory;
    SgExpression * baseexp=new SgVarRefExp(base);

    SgSymbol * franame = s->getSymbolInCurrentScope(mixstring, allocHeapNameOfType(type), NULL);
    SgFunctionRefExp * rafuncexpr = new SgFunctionRefExp (*franame);

    rafuncexpr->addArg(*sizeexp);

    SgExpression * mhexp=new SgVarRefExp(mHandle);
    rafuncexpr->addArg(*mhexp);

    SgStatement *alloc = new SgAssignStmt(*baseexp, *rafuncexpr);

    delete baseexp, rafuncexpr, mhexp;

    return alloc;
}


SgStatement *make_reset_array_base(SgStatement *s,SgSymbol *dad,SgType *type,SgSymbol *base, SgSymbol *mh)
{
  // reset base;
    SgSymbol * symb = s->getSymbolInCurrentScope(mixstring, "reset_array_base", NULL);
    SgCallStmt * funcall = new SgCallStmt (*symb);

    SgExpression * baseexpr=new SgVarRefExp(dad);
    funcall->addArg(*baseexpr);

#ifdef MEM_ALLOC
    SgSymbol * rs= s->getSymbolInCurrentScope(stackNameOfType(type),"mb", NULL);
#else
    SgSymbol * rs= s->getSymbolInCurrentScope(mixstring, stackNameOfType(type), NULL);
#endif
    SgArrayRefExp * stackarray = new SgArrayRefExp(*rs);

    SgExpression * baseexp= new SgVarRefExp(base);
    stackarray->addSubscript(*baseexp);
    funcall->addArg(*stackarray);

    SgExpression *mhExp = new SgVarRefExp(mh);
    funcall->addArg(*mhExp);

    delete baseexpr, stackarray, baseexp;

    return funcall;
}

SgStatement *make_reset_array_base1(SgStatement *s,SgSymbol *dad,SgExpression *address, SgSymbol *mh)
{
  // reset base;
    SgSymbol * symb = s->getSymbolInCurrentScope(mixstring, "reset_array_base", NULL);
    SgCallStmt * funcall = new SgCallStmt (*symb);

    SgExpression * baseexpr=new SgVarRefExp(dad);
    funcall->addArg(*baseexpr);

    funcall->addArg(*address);

    SgExpression *mhExp = new SgVarRefExp(mh);
    funcall->addArg(*mhExp);

    delete baseexpr;

    return funcall;
}

SgStatement *make_reset_in_core_base(SgStatement *s,SgSymbol *dad,SgType *type,SgSymbol *base,SgExpression *mhExp)
{
  // reset base;
    SgSymbol * symb = s->getSymbolInCurrentScope(mixstring, "reset_in_core_base", NULL);
    SgCallStmt * funcall = new SgCallStmt (*symb);

    SgExpression * baseexpr=new SgVarRefExp(dad);
    funcall->addArg(*baseexpr);

#ifdef MEM_ALLOC
    SgSymbol * rs= s->getSymbolInCurrentScope(stackNameOfType(type),"mb", NULL);
#else
    SgSymbol * rs= s->getSymbolInCurrentScope(mixstring, stackNameOfType(type), NULL);
#endif
    SgArrayRefExp * stackarray = new SgArrayRefExp(*rs);

    SgExpression * baseexp= new SgVarRefExp(base);
    stackarray->addSubscript(*baseexp);
    funcall->addArg(*stackarray);

    funcall->addArg(*mhExp);

    delete baseexpr, stackarray, baseexp;

    return funcall;
}

SgExpression *make_glob(SgStatement *s,SgSymbol *irg,SgExpression *index)
{
  // reset base;
    SgSymbol * symb = s->getSymbolInCurrentScope(mixstring, "glob", NULL);
    SgFunctionRefExp * funcall = new SgFunctionRefExp (*symb);

    SgExpression * irgexp=new SgVarRefExp(irg);
    funcall->addArg(*irgexp);

    funcall->addArg(*index);

    delete irgexp;

    return funcall;
}

SgExpression *make_group(SgStatement *s,SgSymbol *dad)
{
  SgSymbol * grp = s->getSymbolInCurrentScope(mixstring, "group", NULL);
  SgFunctionRefExp * grpexp = new SgFunctionRefExp(*grp);

  SgExpression * funref= new SgVarRefExp(dad);
  grpexp->addArg(*funref);
  delete funref;

  return grpexp;
}

SgExpression *make_range(SgStatement *s,SgSymbol *sym,SgExpression *dim)
{
	SgSymbol * frngname = s->getSymbolInCurrentScope(mixstring, "range", NULL);
	SgFunctionRefExp * funcexpr = new SgFunctionRefExp (*frngname);

   	{
	SgExpression * dad = new SgVarRefExp(sym);
	funcexpr->addArg(*dad);
        delete dad;
	}

        funcexpr->addArg(*dim);

	return funcexpr;
}

SgExpression *make_stride(SgStatement *s,SgSymbol *dad,SgExpression *dim)
{
      SgSymbol * str = s->getSymbolInCurrentScope(mixstring, "stride", NULL);
      SgFunctionRefExp * strfun = new SgFunctionRefExp (*str);
	
      {
      SgExpression * dadexp = new SgVarRefExp (dad);
      strfun->addArg(*dadexp);
      delete dadexp;
      }

      strfun->addArg(*dim);

      return strfun;
}

SgExpression *make_loc(SgStatement *s,SgExpression *rngfun,SgExpression *dim)
{
      SgSymbol * loc = s->getSymbolInCurrentScope(mixstring, "loc", NULL);
      SgFunctionRefExp * locfun = new SgFunctionRefExp (*loc);

      locfun->addArg(*rngfun);
      locfun->addArg(*dim);
      return locfun;
}

SgExpression *make_on(SgStatement *s, SgExpression *grpexp)
{
  SgSymbol * on = s->getSymbolInCurrentScope(mixstring, "on", NULL);
  SgFunctionRefExp * onexp = new SgFunctionRefExp(*on);
  
  onexp->addArg(*grpexp);
  return onexp;
}

SgExpression *make_in_core_size(SgStatement *s, SgSymbol *dad )
{
    SgSymbol * fsizename = s->getSymbolInCurrentScope(mixstring, "in_core_size", NULL);
    SgFunctionRefExp * sizefuncexpr = new SgFunctionRefExp (*fsizename);

    {
    SgExpression * dadbaseexp=new SgVarRefExp(dad);
    sizefuncexpr->addArg(*dadbaseexp);
    delete dadbaseexp;
    }

    return sizefuncexpr;
}

SgStatement *make_new_array_section(SgStatement * s, SgSymbol *dadsec, int dim , SgSymbol *dad)
{	
  SgExpression * lhs = new SgVarRefExp (dadsec);
  
  SgSymbol * newsec = s->getSymbolInCurrentScope(mixstring, "new_array_section", NULL);
  SgFunctionRefExp * rhs = new SgFunctionRefExp(*newsec);
  
  SgValueExp * dimexp = new SgValueExp(dim);
  rhs->addArg(*dimexp);
  
  SgExpression * dadexp = new SgVarRefExp (dad);
  rhs->addArg(*dadexp);
  
  SgAssignStmt * assign = new SgAssignStmt(*lhs, *rhs);
  
  delete lhs, rhs, dimexp, dadexp;

  return assign;
}

// define array from array ref like :  a(:,3)
/* input:
 statement: s 
 array ref: arrayexp 
 dad: dadsymb
*/
void define_array_from_arrayref(SgStatement * s, SgArrayRefExp *arrayexp, SgSymbol *dadsymb)
{

  int dimcounter=0;
  SgSymbol * arrayname= arrayexp->symbol();

  for (int i=0; i<arrayexp->numberOfSubscripts(); i++) {
    
    if (!arrayexp->isScalarOnDim(i)) {
        dimcounter++;
        
	SgTripletExp * triplet=(SgTripletExp *)(arrayexp->subscript(i));
	SgExpression *start = triplet->item(0);
	if (start == NULL)  start = get_lbound(arrayname, i);
        analyse_expression(s, start);
	
	SgExpression *end = triplet->item(1);
	if (end == NULL)  end = get_ubound(arrayname, i);
        analyse_expression(s, end);

	SgExpression *stride = triplet->item(2);
	if (stride == NULL)  stride = new SgValueExp(1);
	stride = calculateExp(stride);
        
        SgSymbol * dad =s->getSymbolInCurrentScope(mixstring, "dad", arrayexp->symbol()->identifier());

        SgStatement *funcall = make_set_array_triplet(s, dadsymb, dimcounter,
					            start,end,stride,dad,i+1);
        s->insertStmtBefore(*funcall, *s->controlParent());
        delete funcall;
    }
    else {
        SgExpression * scalar = arrayexp ->subscript(i);
        SgSymbol * dad =s->getSymbolInCurrentScope(mixstring, "dad", arrayexp->symbol()->identifier());

        SgStatement *funcall = make_set_array_scalar(s, dadsymb, scalar, dad, i+1);
        s->insertStmtBefore(*funcall, *s->controlParent());
        delete funcall;
    }
  }
}

// define array from array variable: a
/* input:
 statement: s 
 array symbol: arraysymb 
 dad: dadsymb
*/
void define_array_from_array(SgStatement * s, SgSymbol *arraysymb, SgSymbol *dadsymb)
{
  SgArrayType *arraytype  = isSgArrayType(arraysymb->type());
  ASSERT(arraytype);

  int dimcounter=0;
  for (int i=0; i<arraytype->dimension(); i++) {
    
      dimcounter++;

      SgExpression *start = get_lbound(arraysymb,i);
      analyse_expression(s, start);

      SgExpression *end = get_ubound(arraysymb,i);
      analyse_expression(s, end);

      SgExpression * stride = new SgValueExp(1);
      
      SgSymbol * dad =s->getSymbolInCurrentScope(mixstring, "dad", arraysymb->identifier());

      SgStatement *funcall = make_set_array_triplet(s, dadsymb, dimcounter,
					            start,end,stride,dad,i+1);
      delete stride;

      s->insertStmtBefore(*funcall, *s->controlParent());
      delete funcall;
    }
}

SgStatement *make_set_array_triplet(SgStatement *s,SgSymbol *dadsymb,int dim,SgExpression *start,SgExpression *end,SgExpression *stride,SgSymbol *dad,int dimParent)
{
      SgSymbol * symb= s->getSymbolInCurrentScope(mixstring, "set_array_triplet", NULL);
      SgCallStmt * funcall= new SgCallStmt(*symb);

      {
	SgExpression * dadsecexp = new SgVarRefExp(dadsymb);
	funcall->addArg(*dadsecexp);
	delete dadsecexp;
      }
      {
	SgValueExp * secdim = new SgValueExp (dim);
	funcall->addArg(*secdim);
	delete secdim;
      }
      {
	  funcall->addArg(*start);
	  funcall->addArg(*end);
	  funcall->addArg(*stride);
      }
      {
	SgExpression * dadexp = new SgVarRefExp(dad);
	funcall->addArg(*dadexp);
	delete dadexp;
      }
      {
	SgValueExp * dimExp = new SgValueExp (dimParent);
	funcall->addArg(*dimExp);
	delete dimExp;
      }
      return funcall;
}

SgStatement *make_set_array_scalar(SgStatement *s,SgSymbol *dadsymb,SgExpression *scalar,SgSymbol *dad,int dim)
{
      SgSymbol * symb= s->getSymbolInCurrentScope(mixstring, "set_array_scalar", NULL);
      SgCallStmt * funcall= new SgCallStmt(*symb);
      {
        SgExpression * dadsecexp = new SgVarRefExp(dadsymb);
        funcall->addArg(*dadsecexp);
	delete dadsecexp;
      }
      {
        funcall->addArg(*scalar);
      }
      {
        SgExpression * dadexp = new SgVarRefExp(dad);
        funcall->addArg(*dadexp);
	delete dadexp;
      };
      {
        SgValueExp * dimexp = new SgValueExp (dim);
        funcall->addArg(*dimexp);
	delete dimexp;
      }
      return funcall;
}

SgStatement *make_coef(SgStatement *s,SgSymbol *dad,int dim,SgExpression *l,SgExpression *u,SgExpression *ss,SgExpression *stride,SgExpression *offset,SgExpression *amountExp,int counter,int no)
{
    SgSymbol * coefsym = s->getSymbolInCurrentScope(mixstring, "coef", NULL);
    SgCallStmt * call = new SgCallStmt(*coefsym);

    {
      SgExpression * dadexp = new SgExpression(VAR_REF, NULL, NULL, dad);
      call->addArg(*dadexp);
      delete dadexp;
    };

    {
      SgValueExp * dimexp = new SgValueExp(dim);
      call->addArg(*dimexp);
      delete dimexp;
    }
    {
      call->addArg(*l);
      call->addArg(*u);
      call->addArg(*ss);
    }

    {
      call->addArg(*stride);
      call->addArg(*offset);
    }

    {
      call->addArg(*amountExp);
    }

    {
      SgSymbol * indexsym = s->getSymbolInCurrentScope(mixstring, "u", NULL, counter, no);
      SgExpression * indexexp = new SgExpression(VAR_REF, NULL, NULL, indexsym);
      call->addArg(*indexexp);
      delete indexexp;
    }

    {
      SgSymbol * indexsym = s->getSymbolInCurrentScope(mixstring, "v", NULL, counter, no);
      SgExpression * indexexp = new SgVarRefExp(indexsym);
      call->addArg(*indexexp);
      delete indexexp;
    }

    return call;
}

SgStatement *make_offset(SgStatement *s,SgSymbol *dad,int dim,SgExpression *offset, SgSymbol *v)
{
    SgSymbol * coefsym = s->getSymbolInCurrentScope(mixstring, "offset", NULL);
    SgCallStmt * call = new SgCallStmt(*coefsym);

    {
      SgExpression * dadexp = new SgVarRefExp(dad);
      call->addArg(*dadexp);
      delete dadexp;
    };

    {
      SgValueExp * dimexp = new SgValueExp(dim);
      call->addArg(*dimexp);
      delete dimexp;
    }
    {
      call->addArg(*offset);
    }
    {
      SgExpression * vexp = new SgVarRefExp(v);
      call->addArg(*vexp);
      delete vexp;
    }

    return call;
}

SgStatement *make_same_data_mapping(SgStatement *s,SgSymbol *lhs,SgSymbol *dad1,SgSymbol *dad2)
{
    SgSymbol * funsym = s->getSymbolInCurrentScope(mixstring, "same_data_mapping", NULL);
    SgFunctionCallExp * call = new SgFunctionCallExp(*funsym);

    {
      SgExpression * dadexp = new SgVarRefExp(dad1);
      call->addArg(*dadexp);
      delete dadexp;
    };

    {
      SgExpression * dadexp = new SgVarRefExp(dad2);
      call->addArg(*dadexp);
      delete dadexp;
    };

    SgExpression *lhsExp = new SgVarRefExp(*lhs);
    SgAssignStmt *stmt = new SgAssignStmt(*lhsExp, *call);
    delete lhsExp;
    return stmt;
}

SgStatement * make_index(SgStatement *s,SgSymbol *lhs,SgSymbol *dad, SgType *type)
{
    SgSymbol * funsym= s->getSymbolInCurrentScope(mixstring,nameOfType(type),"index");
    SgFunctionCallExp * call = new SgFunctionCallExp(*funsym);

    {
      SgExpression * dadexp = new SgVarRefExp(dad);
      call->addArg(*dadexp);
      delete dadexp;
    };

    SgExpression *lhsExp = new SgVarRefExp(*lhs);
    SgAssignStmt *stmt = new SgAssignStmt(*lhsExp, *call);
    delete lhsExp;
    return stmt;
}

SgExpression * make_distcode(SgStatement *s,SgSymbol *dad,int dim)
{
	SgSymbol *franame=s->getSymbolInCurrentScope(mixstring,"distcode",NULL);
    	SgFunctionRefExp * rafuncexpr = new SgFunctionRefExp (*franame);
	{
	SgExpression *dadExp = new SgVarRefExp(dad);
    	rafuncexpr->addArg(*dadExp);
	delete dadExp;
	}
	{
	SgExpression *dimExp = new SgValueExp(dim);
    	rafuncexpr->addArg(*dimExp);
	delete dimExp;
	}
	return rafuncexpr;
}

SgStatement *make_clock(SgStatement *s, int flag)
{
    SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "clock", NULL);
    SgCallStmt * afuncall = new SgCallStmt(*fname);
    {
    SgExpression *val0 = new SgValueExp(flag);
    afuncall->addArg(*val0);
    delete val0;
    }
    return afuncall;
}

SgStatement * make_set_global_major(SgStatement *s, int storage)
{
    SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "set_global_major", NULL);
    SgCallStmt * afuncall = new SgCallStmt(*fname);

    {
    SgSymbol *storeSymb = getStorageSymb(s, storage);
    SgExpression *storeExp =new SgVarRefExp(storeSymb);
    afuncall->addArg(*storeExp);
    delete storeExp;
    }
    return afuncall;
}

SgStatement * make_sub_group_grid(SgStatement *s, SgSymbol *subgrp, int rank, SgSymbol *l, SgSymbol *u, SgSymbol *ss, SgSymbol *grp)
{
    SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "sub_group_grid", NULL);
    SgFunctionRefExp * funcexpr = new SgFunctionRefExp (*fname);

    SgExpression *rankExp = new SgValueExp(rank);
    funcexpr->addArg(*rankExp);
    delete rankExp;

    SgExpression *lExp =new SgVarRefExp(l);
    funcexpr->addArg(*lExp);
    delete lExp;

    SgExpression *uExp =new SgVarRefExp(u);
    funcexpr->addArg(*uExp);
    delete uExp;

    SgExpression *sExp =new SgVarRefExp(ss);
    funcexpr->addArg(*sExp);
    delete sExp;

    SgExpression *grpExp =new SgVarRefExp(grp);
    funcexpr->addArg(*grpExp);
    delete grpExp;

    SgExpression *subgrpExp =new SgVarRefExp(subgrp);

    SgStatement *stmt = new SgAssignStmt(*subgrpExp, *funcexpr);
    delete subgrpExp;

    return stmt;
}

SgStatement * make_oc_read(SgStatement *s,SgSymbol *dad,SgExpression * mem_position,SgExpression * start_position,SgSymbol *stride,SgSymbol *number )
{
    SgSymbol * ocreadsym = s->getSymbolInCurrentScope(mixstring, "oc_read", NULL);
    SgCallStmt * call = new SgCallStmt(*ocreadsym);

    {
      SgExpression * dadexp = new SgVarRefExp(dad);
      call->addArg(*dadexp);
      delete dadexp;
    }
    {
      call->addArg(*mem_position);
    }
    {
      call->addArg(*start_position);
    }
    {
      SgExpression * stride_exp = new SgVarRefExp(stride);
      call->addArg(*stride_exp);
      delete stride_exp;
    }
    {
      SgExpression * number_exp = new SgVarRefExp(number);
      call->addArg(*number_exp);
      delete number_exp;
    }
    return call;
}

SgStatement * make_oc_write(SgStatement *s,SgSymbol *dad,SgExpression * mem_position,SgExpression * start_position,SgSymbol *stride,SgSymbol *number )
{
    SgSymbol * ocreadsym = s->getSymbolInCurrentScope(mixstring, "oc_write", NULL);
    SgCallStmt * call = new SgCallStmt(*ocreadsym);

    {
      SgExpression * dadexp = new SgVarRefExp(dad);
      call->addArg(*dadexp);
      delete dadexp;
    }
    {
      call->addArg(*mem_position);
    }
    {
      call->addArg(*start_position);
    }
    {
      SgExpression * stride_exp = new SgVarRefExp(stride);
      call->addArg(*stride_exp);
      delete stride_exp;
    }
    {
      SgExpression * number_exp = new SgVarRefExp(number);
      call->addArg(*number_exp);
      delete number_exp;
    }

    return call;
}

SgStatement * make_oc_prefetch_read(SgStatement *s,SgSymbol *handle,SgSymbol *dad,SgExpression * mem_position,SgExpression * start_position,SgSymbol *stride,SgSymbol *number )
{
    SgSymbol * ocreadsym = s->getSymbolInCurrentScope(mixstring, "oc_prefetch_read", NULL);
    SgFunctionRefExp * call = new SgFunctionRefExp (*ocreadsym);

    {
      SgExpression * dadexp = new SgVarRefExp(dad);
      call->addArg(*dadexp);
      delete dadexp;
    }
    {
      call->addArg(*mem_position);
    }
    {
      call->addArg(*start_position);
    }
    {
      SgExpression * stride_exp = new SgVarRefExp(stride);
      call->addArg(*stride_exp);
      delete stride_exp;
    }
    {
      SgExpression * number_exp = new SgVarRefExp(number);
      call->addArg(*number_exp);
      delete number_exp;
    }

    SgExpression *handle_exp = new SgVarRefExp(handle);
    SgAssignStmt * assignStmt = new SgAssignStmt(*handle_exp, *call);
    return assignStmt;
}

SgStatement * make_prefetch_wait(SgStatement *s,SgSymbol *handle)
{
    SgSymbol *waitsym = s->getSymbolInCurrentScope(mixstring, "prefetch_wait", NULL);
    SgCallStmt * call = new SgCallStmt(*waitsym);
    {
      SgExpression *handle_exp = new SgVarRefExp(handle);
      call->addArg(*handle_exp);
      delete handle_exp;
    }
    return call;
}

SgStatement * make_oc_init(SgStatement *s,SgSymbol *dad,int in_core_size)
{
    SgSymbol * ocinitsym = s->getSymbolInCurrentScope(mixstring, "oc_init", NULL);
    SgCallStmt * call = new SgCallStmt(*ocinitsym);
 
    {
      SgExpression * dadexp = new SgVarRefExp(dad);
      call->addArg(*dadexp);
      delete dadexp;
    }
    {
//      SgValueExp * in_core_exp = new SgValueExp(in_core_size);
      SgExpression *in_core_exp = get_in_core_size(s);
      call->addArg(*in_core_exp);
      delete in_core_exp;
    }

    return call;
}

SgStatement * make_oc_filemap(SgStatement *s,SgSymbol *dad, SgExpression *fileExp, int in_core_size)
{
    SgSymbol * ocinitsym = s->getSymbolInCurrentScope(mixstring, "oc_filemap", NULL);
    SgCallStmt * call = new SgCallStmt(*ocinitsym);
 
    {
      SgExpression * dadexp = new SgVarRefExp(dad);
      call->addArg(*dadexp);
      delete dadexp;
    }
    {
      call->addArg(*fileExp);
    }
    {
      SgSymbol * lenSymb=s->getSymbolInCurrentScope("len");
      SgFunctionCallExp *lenExp=new SgFunctionCallExp(*lenSymb);
      lenExp->addArg(*fileExp);
      call->addArg(*lenExp);
      delete lenExp;
    }
    {
//      SgValueExp * in_core_exp = new SgValueExp(in_core_size);
      SgExpression *in_core_exp = get_in_core_size(s);
      call->addArg(*in_core_exp);
      delete in_core_exp;
    }

    return call;
}

SgStatement * make_set_ic(SgStatement *s,SgSymbol *dad)
{
    SgSymbol * ocinitsym = s->getSymbolInCurrentScope(mixstring, "set_ic", NULL);
    SgCallStmt * call = new SgCallStmt(*ocinitsym);
 
    {
      SgExpression * dadexp = new SgVarRefExp(dad);
      call->addArg(*dadexp);
      delete dadexp;
    }
    return call;
}

SgStatement * make_set_io_process(SgStatement *s, SgExpression *grpExp)
{
    SgSymbol * fsym = s->getSymbolInCurrentScope(mixstring, "set_io_process", NULL);
    SgCallStmt * call = new SgCallStmt(*fsym);

    call->addArg(*grpExp);
    return call;
}
 
// call make_oc_init or make_oc_filemap to initialize a out_of_core array
SgStatement * make_out_of_core(SgStatement *s, SgSymbol *array, SgSymbol *dad, int in_core_size)
{
    if (array->numberOfAttributes(Filemap_Def) == 0) {
	return make_oc_init(s,dad,in_core_size);
    }
    else {
	SgExpression *fileExp = (SgExpression *)array->attributeValue(0, Filemap_Def);
	return make_oc_filemap(s, dad, fileExp, in_core_size);
    }
}

/*  for OUT_OF_CORE */
SgAssignStmt * make_stride_assign(SgStatement *s,SgSymbol * arraySym,char * stride ,int var_num,int i_dim)
{
    SgSymbol * LstrideSym=s->getSymbolInCurrentScope(mixstring,"oc_stride",arraySym->identifier(),var_num);

    SgSymbol * RstrideSym=s->getSymbolInCurrentScope(mixstring, stride,NULL,var_num,i_dim);
    SgExpression * RstrideExp = new SgVarRefExp(RstrideSym);

    SgVariableSymb  *array_name=(SgVariableSymb *)LstrideSym;
    SgArrayRefExp * aryExp=new SgArrayRefExp (*array_name);
    {
    SgExpression * valExp= new SgValueExp(i_dim+1);
    aryExp->addSubscript(*valExp);
    delete valExp;
    }
    SgAssignStmt * assignStmt = new SgAssignStmt(*aryExp, *RstrideExp);
    delete RstrideExp, aryExp;
    return(assignStmt);

}

SgAssignStmt * make_number_assign(SgStatement *s,SgSymbol * arraySym,SgExpression * numberExp,int i_dim )
{
    SgSymbol * LnumberSym=s->getSymbolInCurrentScope(mixstring,"oc_number",arraySym->identifier());
    SgVariableSymb  *array_name=(SgVariableSymb *)LnumberSym;
    SgArrayRefExp * aryExp=new SgArrayRefExp (*array_name);
    SgExpression * valExp= new SgValueExp(i_dim+1);
    aryExp->addSubscript(*valExp);
    delete valExp;
    SgAssignStmt * assignStmt = new SgAssignStmt(*aryExp, *numberExp);
    delete aryExp;
    return(assignStmt);
}


/*****************************************************************************
			FACILITIES
*****************************************************************************/

Distribute_def * getDistributeDef(SgSymbol *arraysym)
{
    if (is_dynamic(arraysym) || is_pointer(arraysym) || is_allocatable(arraysym)) {
	return NULL;
    }

   if (arraysym->numberOfAttributes(Align_Def) == 0)
   { 	// dummy has only distribute
        if (arraysym->numberOfAttributes(Dist_Def)==0) return NULL;   // INHERIT,not sure
        Distribute_def * ddef = (Distribute_def *)arraysym->attributeValue(0, Dist_Def);
	return ddef;
   }
   else {
        Alignment_def * adef = (Alignment_def *) arraysym->attributeValue(0, Align_Def);
        SgSymbol * tsym = adef->gettosymbol();
	if (tsym->numberOfAttributes(Dist_Def) == 0) return NULL;
        Distribute_def * ddef = (Distribute_def *) tsym->attributeValue(0, Dist_Def);
	return ddef;
   }
   return NULL;
}

// if P(1) , return 1
int one_processor(Processor_def *pdef)
{
     // only one processor
   if (pdef->getDim() == 1) {
	int procnum=0;
	SgExpression *lb = pdef->getStart(0);
	SgExpression *ub = pdef->getEnd(0);
	if (lb->isInteger() && ub->isInteger())
		procnum = ub->valueInteger() - lb->valueInteger() +1;
	if (procnum == 1) return 1;
   }
   return 0;
}

// check if an symbol's dimension: dim is collapsed
// 1: is dim-collapsed
// 0: not
//-1: not sure
int array_dim_collapsed(SgSymbol *arraysym,int dim)
{
    if (is_dynamic(arraysym) || is_pointer(arraysym) || is_allocatable(arraysym)) {
	return -1;
    }

   if (arraysym->numberOfAttributes(Align_Def) == 0)
   { 	// dummy has only distribute
        if (arraysym->numberOfAttributes(Dist_Def)==0) return -1;   // INHERIT,not sure
        Distribute_def * ddef = (Distribute_def *)arraysym->attributeValue(0, Dist_Def);
	ASSERT(dim < ddef->getDim());

	Processor_def * pdef = (Processor_def *)ddef->gettosymbol()->attributeValue(0, Proc_Def);
	if (one_processor(pdef) == 1) return 1;

   	if (ddef->distCode(dim)==NOP) return 1;
	else return 0;
   }
   else {
        Alignment_def * adef = (Alignment_def *) arraysym->attributeValue(0, Align_Def);
        SgSymbol * tsym = adef->gettosymbol();
        if (tsym->numberOfAttributes(Dist_Def)==0) return -1;	// align to INHERIT, so not sure
        Distribute_def * ddef = (Distribute_def *) tsym->attributeValue(0, Dist_Def);

	if (is_template(tsym))
	{
	  Processor_def * pdef = (Processor_def *)ddef->gettosymbol()->attributeValue(0, Proc_Def);
	  if (one_processor(pdef) == 1) return 1;
	}

	ASSERT(dim < adef->getDim());
        int alignwith = adef->getAlignPair(dim);
   	if (ddef->distCode(alignwith)==NOP) return 1;
	else return 0;
   }
   return 0;
}

// 1: if array is all collapsed
// 0: not
//-1: not sure
int array_all_collapsed(SgSymbol *array){
    int i;
    int all_collapsed = 0;

    if (is_dynamic(array) || is_pointer(array) || is_allocatable(array)) {
        all_collapsed = -1;
	return all_collapsed;
    }

    if (array->numberOfAttributes(Align_Def)==0) {
        if (array->numberOfAttributes(Dist_Def)==0) return -1;	// INHERIT, so not sure
        Distribute_def * ddef = (Distribute_def *)array->attributeValue(0, Dist_Def);

	Processor_def * pdef = (Processor_def *)ddef->gettosymbol()->attributeValue(0, Proc_Def);
	if (one_processor(pdef) == 1) return 1;

	for (i=0; i<ddef->getDim(); i++) {
     		if (ddef->distCode(i)!=NOP) return 0;
	}
        all_collapsed = 1;
    }
    else {
    	Alignment_def *adef = (Alignment_def *)array->attributeValue(0,Align_Def);

	SgSymbol * temp = adef->gettosymbol();
        if (temp->numberOfAttributes(Dist_Def)==0) return -1;	// align to INHERIT, so not sure
	Distribute_def * ddef = (Distribute_def *)temp->attributeValue(0, Dist_Def);

	if (is_template(temp))
	{
	   Processor_def * pdef = (Processor_def *)ddef->gettosymbol()->attributeValue(0, Proc_Def);
	   if (one_processor(pdef) == 1) return 1;
	}

	for (i=0; i<adef->getDim(); i++) {
     		int templateDim=adef->getAlignPair(i);
     		if (ddef->distCode(templateDim)!=NOP) return 0;
	}
	all_collapsed = 1;

   }

/*
   // if array is a dummy, reduced dimension may cause communication.!
   // although this array is collapsed, the parent array maynot be all-collapsed.
   if (all_collapsed == 1 && is_dummy(array)) return -1;
*/

   return all_collapsed;
}

// if symbol is dummy ,return 1
int is_dummy(SgStatement *s, SgSymbol *symb)
{
  SgStatement *scope = s->getScopeForDeclare();
  SgFunctionSymb *funcName = isSgFunctionSymb(scope->symbol());
  if (funcName == NULL) return 0;
  
  int dummy=0;
  for (int i=0; i<funcName->numberOfParameters(); i++)
	if (strcasecmp(symb->identifier(), funcName->parameter(i)->identifier()) == 0) 
		{ dummy=1; break; }
  return dummy;
}

// if intrinsic func is deal as function , instead of change to subroutine()
// return 1
int function_intrinsic(SgSymbol *func)
{
   if (strcasecmp(func->identifier(), "number_of_processors") == 0) return 1;
   return 0;
}


int is_out_of_core(SgSymbol *symb)
{
   return symb->numberOfAttributes(Outofcore_Def) > 0 ? 1:0;
}

int is_dynamic(SgSymbol *symb)
{
   return symb->numberOfAttributes(Dynamic_Def) > 0 ? 1:0;
}

int is_inherit(SgSymbol *symb)
{
   return symb->numberOfAttributes(Inherit_Def) > 0 ? 1:0;
}

int is_allocatable(SgSymbol *symb)
{
   return symb->attributes() &  ALLOCATABLE_BIT ? 1:0;
}

int is_pointer(SgSymbol *symb)
{
   return symb->attributes() &  POINTER_BIT ? 1:0;
}

int unknown_address(SgSymbol *symb)
{
   if (is_dynamic(symb) || is_allocatable(symb) || is_pointer(symb)) return 1;
   if (symb->numberOfAttributes(UNBOUND_ARRAY) > 0) return 1;
   return 0;
}

int is_resident(SgSymbol *symb)
{
   return symb->numberOfAttributes(Resident_Def) > 0 ? 1:0;
}

int is_template(SgSymbol *symb)
{
   return symb->type()->variant() == T_TEMPLATE;
}

int is_common(SgSymbol *symb)
{
   return symb->numberOfAttributes(Common_Def) > 0 ? 1:0;
}

SgStatement *get_contains(SgStatement *s, int &no)
{
   int variant = s->variant();
   ASSERT(variant == PROG_HEDR || variant == PROC_HEDR || variant == FUNC_HEDR  || variant == MODULE_STMT);
   for (no=0; no<s->numberOfChildrenList1(); no++)
   {
	SgStatement *stmt = s->childList1(no);
	if (stmt->variant() == CONTAINS_STMT) return stmt;
   }
   return NULL;
}

int getArrayStorage(SgSymbol *array)
{
      if (array->numberOfAttributes(Major_Def) > 0) {
	return (int)array->attributeValue(0,Major_Def);
      }
      else
	  // use default storage mode 
	return storageMode;
}

SgSymbol *getStorageSymb(SgStatement *s, int storage)
{
      switch (storage) {
      case COLUMN_MAJOR:
    	return  s->getSymbolInCurrentScope("COLUMN_MAJOR"); 
      case ROW_MAJOR:
    	return  s->getSymbolInCurrentScope("ROW_MAJOR"); 
      default:
	ASSERT(0);
      }
      return NULL;
}

//  change array  A(pcrc_sdx0) => int_mb(pcrc_tbs_a + pcrc_sdx0)
//  when A is allocatable , dynamic
//  call pcrc_integer_index before call.
void change_array_by_stack(SgStatement *s, SgArrayRefExp *array)
{
      SgSymbol *arraysym = array->symbol();
      SgSymbol * basesym = s->getSymbolInCurrentScope(mixstring, "tbs", arraysym->identifier());
      SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arraysym->identifier());
      SgStatement * index_stmt = make_index(s, basesym, dad, arraysym->type());
      s->insertStmtBefore(*index_stmt, *s->controlParent());
      SgExpression *sub = array->subscript(0);
      SgExpression *newsub = &( *sub->copyPtr() + *new SgVarRefExp(basesym) );
      newsub = calculateExp(newsub);
      sub->replace( *newsub );
      SgSymbol * stack = s->getSymbolInCurrentScope(stackNameOfType(arraysym->type()),"mb", NULL); 
      array->setSymbol(*stack);
}

// change scalar variable by stack
void change_variable_by_stack(SgStatement *s, SgExpression *var)
{
      SgSymbol *sym = var->symbol();
      SgSymbol * basesym = s->getSymbolInCurrentScope(mixstring, "tbs", sym->identifier());
      SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", sym->identifier());
      SgStatement * index_stmt = make_index(s, basesym, dad, sym->type());
      s->insertStmtBefore(*index_stmt, *s->controlParent());

      SgSymbol * stack = s->getSymbolInCurrentScope(stackNameOfType(sym->type()),"mb", NULL); 
      SgArrayRefExp *array = new SgArrayRefExp(*stack);
      SgExpression *subs = new SgVarRefExp(basesym);
      array->addSubscript(*subs);
      var->replace(*array);
}


/*****************************************************************************
			PROCESSOR GROUP
*****************************************************************************/

// check if proc1 is sub group of proc2
int is_sub_group(SgExpression *proc1, SgExpression *proc2)
{
      SgSymbol *sym1 = proc1->symbol();
      SgSymbol *sym2 = proc2->symbol();
      if ( strcasecmp(sym1->identifier(), sym2->identifier()) != 0)
	return 0;
	// proc2 is whole processor group
      SgSymbol *procname = sym1;
      Processor_def * pdef = (Processor_def *)procname->attributeValue(0, Proc_Def);
      if ( !isSgArrayRefExp(proc2) ) return 1;
	// proc1 is whole processor group
      if (isSgArrayRefExp(proc1) ) return 0;
      SgArrayRefExp *sub1 = isSgArrayRefExp(proc1);
      SgArrayRefExp *sub2 = isSgArrayRefExp(proc2);
      ASSERT(sub1->numberOfSubscripts() == sub2->numberOfSubscripts());
      for (int i=0; i<sub1->numberOfSubscripts(); i++)
      {
	SgExpression *lb1 = get_lbound_from_arrayexp(sub1, i);
	SgExpression *ub1 = get_ubound_from_arrayexp(sub1, i);
	SgExpression *lb2 = get_lbound_from_arrayexp(sub2, i);
	SgExpression *ub2 = get_ubound_from_arrayexp(sub2, i);
          // compare if equal
	if ( *lb1 == *lb2 && *ub1 == *ub2 ) continue;
	if (lb1->isInteger() && lb2->isInteger() &&
	    ub1->isInteger() && ub2->isInteger() ) {
	  int l1 = lb1->isInteger();
	  int l2 = lb2->isInteger();
	  if (l1 < l2) return 0;
	  int u1 = ub1->isInteger();
	  int u2 = ub2->isInteger();
	  if (u1 > u2) return 0;
	  continue;
	}
	else
	  return 0;
      }
      return 1;
}


int sub_processor(SgSymbol *sym1, SgSymbol *sym2)
{
      Distribute_def * ddef1 = getDistributeDef(sym1);
      Distribute_def * ddef2 = getDistributeDef(sym2);
      if (ddef1 == NULL || ddef2 == NULL) return 0;
      int isSub = is_sub_group(ddef1->getProcs(), ddef2->getProcs());
if (isSub == 0)
printf(" %s is not the sub group of %s\n", sym1->identifier(), sym2->identifier());
      return isSub;
}

// check if proc1 is sub group of proc2
int is_same_group(SgExpression *proc1, SgExpression *proc2)
{
      SgSymbol *sym1 = proc1->symbol();
      SgSymbol *sym2 = proc2->symbol();
/*	assume all the whole processor group in program is same
      if ( strcasecmp(sym1->identifier(), sym2->identifier()) != 0)
	return 0;
*/

      int isSub1 = isSgArrayRefExp(proc1)?1:0;
      int isSub2 = isSgArrayRefExp(proc2)?1:0;
      if (isSub1 == 0 && isSub2 == 0)
	return 1;
      if ( isSub1 * isSub2 == 0 )	// one is sub, and another is not
	return 0;

	// proc2 is whole processor group
      SgSymbol *procname = sym1;
      Processor_def * pdef = (Processor_def *)procname->attributeValue(0, Proc_Def);
      SgArrayRefExp *sub1 = isSgArrayRefExp(proc1);
      SgArrayRefExp *sub2 = isSgArrayRefExp(proc2);
      ASSERT(sub1->numberOfSubscripts() == sub2->numberOfSubscripts());

      for (int i=0; i<sub1->numberOfSubscripts(); i++)
      {
	SgExpression *lb1 = get_lbound_from_arrayexp(sub1, i);
	SgExpression *ub1 = get_ubound_from_arrayexp(sub1, i);
	SgExpression *lb2 = get_lbound_from_arrayexp(sub2, i);
	SgExpression *ub2 = get_ubound_from_arrayexp(sub2, i);
          // compare if equal
	if ( *lb1 == *lb2 && *ub1 == *ub2 ) continue;
	return 0;
      }
      return 1;
}


// sym1 & sym2 is array name
// 0: not 1: yes -1:not sure
int same_processor_group(SgSymbol *sym1, SgSymbol *sym2)
{
      Distribute_def * ddef1 = getDistributeDef(sym1);
      Distribute_def * ddef2 = getDistributeDef(sym2);
      if (ddef1 == NULL || ddef2 == NULL) return -1;
      int isSame = is_same_group(ddef1->getProcs(), ddef2->getProcs());
if (isSame == 0)
printf("Info: array %s and %s have not same processor group.\n", sym1->identifier(), sym2->identifier());
else if (isSame == -1)
printf("Info: array %s and %s have unknown same processor group.\n", sym1->identifier(), sym2->identifier());
      return isSame;
}


